# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import collections
import itertools
import logging
import time

import requests
from django.core.handlers.base import BaseHandler
from django.test.client import RequestFactory
from requests.adapters import BaseAdapter
from requests.cookies import extract_cookies_to_jar
from requests.exceptions import ConnectionError, Timeout
from requests.models import RequestEncodingMixin, Response
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers

from rest_models.backend.exceptions import FakeDatabaseDbAPI2
from rest_models.backend.utils import message_from_response

logger = logging.getLogger("django.db.backends")


def build_url(url, params):
    """
    only for display purpose, parse the params and url to build the final api
    :param str url: the url path to use
    :param dict params: the dict with the GET parameters, as accepted by requests
    :return: the url with the get part
    :rtype: str
    """
    if params is None:
        result = url
    else:
        result = "%s?%s" % (url, RequestEncodingMixin._encode_params(params))
    return result


class LocalApiAdapter(BaseAdapter):

    SPECIAL_URL = "http://localapi"

    def __init__(self):
        self.request_factory = RequestFactory()
        super(LocalApiAdapter, self).__init__()

    def prepared_request_to_wsgi_request(self, prepared_request):
        """
        transform a PreparedRequest into a WsgiRequest for django to use
        :param requests.models.PreparedRequest prepared_request:
        :return: the request ready to use for django
        :rtype: django.core.handlers.wsgi.WSGIRequest
        """
        wsgi_request = self.request_factory.generic(
            method=prepared_request.method,
            path=prepared_request.url,
            data=prepared_request.body,
            content_type=prepared_request.headers.get('Content-Type', 'application/x-www-form-urlencoded')
        )
        for name, val in prepared_request.headers.items():
            wsgi_request.META['HTTP_' + name.upper()] = val
        return wsgi_request

    def http_response_to_response(self, http_response, prepared_request):
        """
        transform a WSGIResponse into a requests's Response model
        :param django.http.response.HttpResponse http_response: the http response send by django view
        :return: the requests's Response model corresponding to the http_response
        :rtype: Response
        """
        response = Response()

        # Fallback to None if there's no status_code, for whatever reason.
        response.status_code = getattr(http_response, 'status_code', None)

        # Make headers case-insensitive.
        response.headers = CaseInsensitiveDict(getattr(http_response._headers, 'headers', {}))

        # Set encoding.
        response.encoding = get_encoding_from_headers(response.headers)
        response.raw = http_response
        response.reason = response.raw.reason_phrase
        response._content = http_response.content
        req = prepared_request

        if isinstance(req.url, bytes):  # pragma: no cover
            response.url = req.url.decode('utf-8')
        else:
            response.url = req.url

        # Add new cookies from the server.
        extract_cookies_to_jar(response.cookies, req, response)

        # Give the Response some context.
        response.request = req
        response.connection = self

        return response

    def send(self, request, stream=False, timeout=None, verify=True,
             cert=None, proxies=None):
        handler = BaseHandler()
        handler.load_middleware()
        wsgi_request = self.prepared_request_to_wsgi_request(request)
        http_response = handler.get_response(wsgi_request)
        requests_response = self.http_response_to_response(http_response, request)
        return requests_response

    def close(self):  # pragma: no cover
        pass


class ApiVerbShortcutMixin(object):

    def get(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("get", url, params=params, json=json, **kwargs)

    def head(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("head", url, params=params, json=json, **kwargs)

    def post(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("post", url, params=params, json=json, **kwargs)

    def put(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("put", url, params=params, json=json, **kwargs)

    def patch(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("patch", url, params=params, json=json, **kwargs)

    def options(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("options", url, params=params, json=json, **kwargs)

    def delete(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        """
        return self.request("delete", url, params=params, json=json, **kwargs)


class DebugApiConnectionWrapper(ApiVerbShortcutMixin):
    def __init__(self, connection, db):
        self.connection = connection
        self.db = db

    def __getattr__(self, attr):  # pragma: no cover
        cursor_attr = getattr(self.connection, attr)
        return cursor_attr

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.connection.close()

    def request(self, method, url, **kwargs):

        start = time.time()
        try:
            return self.connection.request(method, url, **kwargs)
        finally:
            stop = time.time()
            duration = stop - start
            sql = build_url(url, kwargs['params'])
            self.db.queries_log.append({
                'sql': "%s %s ||| %s" % (method, sql, kwargs),
                'time': "%.3f" % duration,
            })
            logger.debug('(%.3f) %s %s; args=%s' % (duration, method, sql, kwargs),
                         extra={'duration': duration, 'sql': sql, 'params': kwargs, 'method': method}
                         )


class ApiConnexion(ApiVerbShortcutMixin):
    """
    wrapper for request.Session that in fact implement useless methods like rollback which
    is not possible with a rest API
    """
    def __init__(self, url, auth=None, retry=3, timeout=3, backend=None, middlewares=(), ssl_verify=None):
        """
        create a persistent connection to the api
        :param str url: the base url for the api (host + port + start path)
        :param ApiAuthBase auth: the auth provider
        :param int retry: number of retry
        :param int|(int, int) timeout: the timeout to pass to the api
        :param DatabaseWrapper backend: the backend
        :param list[Middleware]|tuple[Middleware] middlewares:the list of middleware to execute for each query.
        """
        self.session = requests.Session()
        self.session.mount(LocalApiAdapter.SPECIAL_URL, LocalApiAdapter())
        self.session.auth = self.auth = auth
        self.url = url
        self.retry = retry
        self.timeout = timeout
        self.backend = backend
        self._middlewares_scheduler = collections.defaultdict(list)
        self._requestid = 0
        if ssl_verify is not None:
            self.session.verify = ssl_verify
        for middleware in middlewares:
            self.push_middleware(middleware, 8)

    @property
    def middlewares(self):
        """
        the list of middleware to iterate over. ordered by priority from 1 to 10
        :return:
        """
        return list(itertools.chain(*(v for k, v in sorted(self._middlewares_scheduler.items()))))

    def push_middleware(self, middleware, priority=5):
        """
        push a middleware into the list, which will be called at each request and response processing
        :param ApiMiddleware middleware: the middleware to apppend
        :return:
        """
        self._middlewares_scheduler[priority].insert(0, middleware)

    def pop_middleware(self, middleware):
        for middlewares in self._middlewares_scheduler.values():
            try:
                middlewares.remove(middleware)
            except ValueError:
                pass

    def close(self):
        self.session.close()

    def __exit__(self, *args):
        self.close()

    def __enter__(self):
        return self

    def inc_request_id(self):
        """
        increment the request id and then return it
        :return:
        """
        self._requestid += 1
        return self._requestid

    def execute(self, sql, params=None):
        """

        :param sql: the usless sql
        :param params: the userless params
        :return:
        """
        params = params or {}
        return self._make_request(params)

    def _make_request(self, params):
        """
        finaly make the request. will pass all middleware on send and recieve
        :param params:
        :return:
        """
        requestid = self.inc_request_id()
        middlewares = self.middlewares
        i = 0
        for i, middleware in enumerate(middlewares):
            response = middleware.process_request(params, requestid, self)
            if response is not None:
                break
        else:
            # if the middleware did not override the real response, we make the query
            response = self.session.request(**params)
        for middleware in middlewares[i::-1]:  # iterate over all previously executed middlewares
            response = middleware.process_response(params, response, requestid) or response
        return response

    def rollback(self):  # pragma: no cover
        pass

    def commit(self):  # pragma: no cover
        pass

    def get_timeout(self):
        return self.timeout

    def request(self, method, url, **kwargs):
        """
        wrapper for requests.Session.get

        :param unicode method: the method to use
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes data: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param float|tuple[float, float] timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :rtype: requests.Response

        """
        kwargs.setdefault("allow_redirects", False)
        kwargs.setdefault("timeout", self.get_timeout())
        kwargs.setdefault('stream', False)

        assert not url.startswith("/"), "the url should not start with a «/»"
        if url != '' and not self.url.endswith('/'):
            url = '/' + url
        error = 0
        last_exception = None

        while error <= self.retry:
            try:
                real_url = self.url + url
                # to stay compatible with django_debug_toolbar, we must
                # call execute on the cursor return by the backend, since this one is replaced
                # by django-debug-toolbar to state the query.
                if self.backend:
                    execute = self.backend.cursor().execute
                else:
                    execute = self.execute
                response = execute("%s %s" % (method.upper(), real_url), dict(method=method, url=real_url, **kwargs))

            except Timeout as e:
                error += 1
                last_exception = e
            except ConnectionError as e:
                error += 1
                last_exception = e
            else:

                if self.auth and hasattr(self.auth, 'raise_on_response_forbidden'):
                    self.auth.raise_on_response_forbidden(response)
                else:
                    if response.status_code in (403, 401):
                        raise FakeDatabaseDbAPI2.ProgrammingError(
                            "Access to database is Forbidden for user %s.\n%s" %
                            (self.auth[0] if isinstance(self.auth, tuple) else self.auth,
                             message_from_response(response))
                        )
                return response
        raise FakeDatabaseDbAPI2.OperationalError(
            "cound not connect to server: %s\nIs the API running on %s ? tried %d times" %
            (last_exception, self.url, error))



import os
import subprocess

from django.db.backends.base.client import BaseDatabaseClient
from django.test.testcases import LiveServerThread, _StaticFilesHandler

from rest_models.backend.connexion import LocalApiAdapter


class DatabaseClient(BaseDatabaseClient):
    executable_name = 'bash'
    port_range = (8097, 9015)

    def start_server_thread(self):
        self.server_thread = LiveServerThread('localhost', range(*self.port_range),
                                              _StaticFilesHandler)
        self.server_thread.daemon = True
        self.server_thread.start()

        # Wait for the live server to be ready
        self.server_thread.is_ready.wait()
        if self.server_thread.error:
            # Clean up behind ourselves, since tearDownClass won't get called in
            # case of errors.
            self.stop_server_thread()
            raise self.server_thread.error
        return 'http://%s:%s' % (
                self.server_thread.host, self.server_thread.port)

    def stop_server_thread(self):
        # There may not be a 'server_thread' attribute if setUpClass() for some
        # reasons has raised an exception.
        if hasattr(self, 'server_thread'):
            # Terminate the live server's thread
            self.server_thread.terminate()
            self.server_thread.join()

    def runshell(self):
        resty_path = os.path.join(os.path.dirname(__file__), "exec", "resty")
        args = [self.executable_name, "--init-file", resty_path]

        cname = self.connection.settings_dict['NAME']
        if cname.startswith(LocalApiAdapter.SPECIAL_URL):
            cname = cname.replace(LocalApiAdapter.SPECIAL_URL, self.start_server_thread())
        cname = cname.rstrip("/") + "*"
        envs = os.environ.copy()
        envs.update(dict(
            _EXTRA_CURL_AUTH=self.get_middleware_curl_args(),
            _resty_host=cname)
        )
        self.execute_subprocess(args=args, env=envs)

        self.stop_server_thread()

    def execute_subprocess(self, args, env):
        subprocess.call(args, env=env)  # pragma: no cover

    def get_middleware_curl_args(self):
        """
        return the curl extra args to authorize ourselve to the api
        :return:
        """
        return ""



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import datetime
import logging
from collections import namedtuple

from django.db.utils import ProgrammingError
from requests.auth import AuthBase, HTTPBasicAuth

from rest_models.backend.exceptions import FakeDatabaseDbAPI2
from rest_models.backend.utils import message_from_response

try:
    from urllib.parse import urlparse, urlunparse
except ImportError:  # pragma: no cover
    from urlparse import urlparse, urlunparse

logger = logging.getLogger(__name__)


Token = namedtuple('Token', 'expiredate access_token token_type scope')


class ApiAuthBase(AuthBase):
    def __init__(self, databasewrapper, settings_dict):
        """
        create the auth backend for the databasewrapper and settings
        :param rest_models.backend.base.DatabaseWrapper databasewrapper: the database wrapper
        :param dict settings_dict: the settings for this connection
        """
        self.settings_dict = settings_dict
        self.databasewrapper = databasewrapper

    def raise_on_response_forbidden(self, response):
        if response.status_code == 403:
            raise FakeDatabaseDbAPI2.ProgrammingError(
                "Access to database is Forbidden for user %s on %s.\n%s" % (
                    self.settings_dict['USER'],
                    self.settings_dict['NAME'],
                    message_from_response(response)
                )
            )


class BasicAuth(ApiAuthBase):

    def __init__(self, databasewrapper, settings_dict):
        self.backend = HTTPBasicAuth(settings_dict['USER'], settings_dict['PASSWORD'])
        super(BasicAuth, self).__init__(databasewrapper, settings_dict)

    def __call__(self, request):
        return self.backend(request)


class OAuthToken(ApiAuthBase):

    @property
    def token(self):
        """
        return the valide token or fetch one new
        :rtype: Token
        """
        if not hasattr(self, '_token'):
            token = self._token = self.get_token()
        else:
            token = self._token
        if self.has_expired(token):
            token = self._token = self.get_token()
        return token

    @property
    def url_token(self):
        url_token = self.settings_dict.get('OPTIONS', {}).get('OAUTH_URL', '/oauth2/token/')
        parsed_url = urlparse(url_token)
        if parsed_url.netloc == '':
            name_url = urlparse(self.settings_dict['NAME'])
            url_token = urlunparse(name_url[:2] + parsed_url[2:])
        return url_token

    def get_token(self):
        """
        query the api to retrive a OAuth token
        :return: the token from the api
        :rtype: Token
        """
        conn = self.databasewrapper.cursor()
        params = {'grant_type': 'client_credentials'}
        # Get client credentials params
        response = conn.session.request(
            'POST',
            self.url_token,
            params=params,
            auth=(self.settings_dict['USER'], self.settings_dict['PASSWORD']),
            stream=False
        )
        if response.status_code != 200:
            raise ProgrammingError("unable to retrive the oauth token from %s: %s" %
                                   (self.url_token,
                                    message_from_response(response))
                                   )
        data = response.json()
        return Token(
            datetime.datetime.now() + datetime.timedelta(seconds=data['expires_in']),
            data['access_token'],
            data['token_type'],
            data['scope']
        )

    def has_expired(self, token):
        """
        check if a token has expired
        :param Token token: the token to check
        :return: True if the token has expired
        :rtype: bool
        """
        # false if expired in 10 sec
        return token.expiredate < (datetime.datetime.now() + datetime.timedelta(seconds=10))

    def __call__(self, request):
        if request.url != self.url_token:
            request.headers['Authorization'] = "Bearer %s" % self.token.access_token
        return request



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging
from importlib import import_module

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.validation import BaseDatabaseValidation

from rest_models.backend.connexion import ApiConnexion, DebugApiConnectionWrapper
from rest_models.backend.exceptions import FakeDatabaseDbAPI2

from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor

logger = logging.getLogger(__name__)


def import_class(path):
    """
    import a component of a module by his path.
    ie module.submodule.ClassName return the class ClassName
    don't work for nested item. the class must be on the root of the module
    :param str path: the path to import
    :return: the class
    :rtype: type
    """
    lpath = path.split(".")
    module = import_module(".".join(lpath[:-1]))
    obj = getattr(module, lpath[-1])
    return obj


class DatabaseWrapper(BaseDatabaseWrapper):

    Database = FakeDatabaseDbAPI2
    vendor = 'rest_api'
    SchemaEditorClass = DatabaseSchemaEditor

    def __init__(self, *args, **kwargs):
        self.connection = None  # type: ApiConnexion

        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        self.features = DatabaseFeatures(self)
        self.ops = DatabaseOperations(self)
        self.client = DatabaseClient(self)
        self.creation = DatabaseCreation(self)
        self.introspection = DatabaseIntrospection(self)
        self.validation = BaseDatabaseValidation(self)

    def get_connection_params(self):
        authpath = self.settings_dict.get('AUTH', None)
        auth = authpath and import_class(authpath)(self, self.settings_dict)
        options = self.settings_dict.get('OPTIONS', {})

        params = {
            'url': self.settings_dict['NAME'],
            'auth': auth,
            'timeout': self.timeout,
            'backend': self,
            'middlewares': [import_class(path)() for path in self.settings_dict.get('MIDDLEWARES', ())],
            'ssl_verify': options.get('SSL_VERIFY', True)
        }
        return params

    def get_new_connection(self, conn_params):
        return ApiConnexion(**conn_params)

    @property
    def timeout(self):
        return self.settings_dict['OPTIONS'].get('TIMEOUT', 10)

    def init_connection_state(self):
        self.autocommit = True

    def create_cursor(self):
        return self.connection

    def make_cursor(self, cursor):
        return cursor

    def make_debug_cursor(self, cursor):
        return DebugApiConnectionWrapper(cursor, self)

    def close(self):
        # do nothing
        pass

    def _start_transaction_under_autocommit(self):
        pass

    def is_usable(self):
        c = self.connection
        try:
            c.head('', timeout=self.timeout)
            return True
        except FakeDatabaseDbAPI2.OperationalError:
            return False

    def _set_autocommit(self, autocommit):
        pass

    def cursor(self):
        # type: () -> ApiConnexion
        return super(DatabaseWrapper, self).cursor()

    def check(self):
        from rest_models.checks import api_struct_check

        api_struct_check(None)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from contextlib import contextmanager

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.test.testcases import TestCase

from rest_models.backend.middlewares import ApiMiddleware
from rest_models.router import get_default_api_database
from rest_models.utils import JsonFixtures, dict_contains


def not_found_raise(url, middleware):
    """
    raise an exception if the query is not found
    :param str url: the asked url
    :param MockDataApiMiddleware middleware: the middleware
    :return:
    """
    raise Exception("the query %r was not provided as mocked data" % url)


def not_found_continue(url, middleware):
    """
    do not do anything
    :param str url: the asked url
    :param MockDataApiMiddleware middleware: the middleware
    :return:
    """
    return None  # returning None in a process_request mean continue


class MockDataApiMiddleware(ApiMiddleware):
    """
    a middleware that totaly disablow a real query on the api.
    it will return a data for a given url, and record the numbers of query
    made.

    data_for_url is a dict with:

    .. code-block:: python

        {"/model/1/": [
            {
                'filter': {
                    'params': {
                        'filter{id.in}': [15894],
                        'exclude[]': {'*'}
                    }
                },
                'data': {...} ## copy past from the api for this response.
        ]
        }
    """

    def __init__(self, data_for_url=None, not_found=not_found_raise):
        data_for_url = data_for_url or {}
        self.data_for_url = data_for_url
        self.not_found = not_found

    def process_request(self, params, requestid, connection):
        if not params['url'].startswith(connection.url):
            raise ImproperlyConfigured("strage case where the query don't go to our api")  # pragma: no cover
        url = params['url'][len(connection.url):]

        try:
            results_for_url = self.data_for_url[url]
        except KeyError:
            return self.not_found(url, self)
        # we have many results for this url.
        # the mocked result can add a special «filter» value along with «data»
        # that all items must match the one in the params to be ok.
        result_found = None
        for mocked_result in results_for_url:
            filters = mocked_result.get('filter', [{}])
            if not isinstance(filters, list):
                filters = [filters]
            for filter in filters:
                if dict_contains(filter, params):
                    # all item in the for is ok.
                    # we break and so won't pass into the «else» of the first for
                    result_found = mocked_result
                    break
            if result_found:
                break
        if not result_found:
            # no mocked data have matched
            return self.not_found(url, self)

        data = result_found.get('data')
        status_code = result_found.get('status_code')
        if data is None and not status_code:
            # None return 204
            return self.empty_response()
        elif data is None:
            # int: statuscode
            return self.make_response(None, status_code)
        elif isinstance(data, (dict, list)):
            # dict,list: data to return
            return self.data_response(data, status_code)
        else:
            raise Exception("the given data don't match a proper type : %s not in (int, dict, list, None)" %
                            type(data))


class RestModelTestMixin(object):
    """
    a test case mixin that add the feathure to mock the response from an api

    """
    rest_fixtures = {}
    """
    fixtures to give for this tests. all missing values will trigger a exception
    """

    database_rest_fixtures = None

    @classmethod
    def setUpClass(cls):
        # populate the database_rest_fixtures from the rest_fixtures for moste case where
        # there is only one database targeting an api.
        super(RestModelTestMixin, cls).setUpClass()
        if cls.database_rest_fixtures is None:
            cls.database_rest_fixtures = {
                get_default_api_database(settings.DATABASES): cls.rest_fixtures
            }

    def setUp(self):
        # add all mock_data middleware to the databsase
        self._mock_data_middleware = {}
        self.rest_fixtures_variables = {}  # should be update by the tests, and by side effects work on mocked data
        for db_name, fixtures in self.database_rest_fixtures.items():
            fixtures = JsonFixtures(fixtures)
            fixtures.set_variable(self.rest_fixtures_variables)
            self._mock_data_middleware[db_name] = MockDataApiMiddleware(fixtures)
            dbwrapper = connections[db_name]
            dbwrapper.cursor().push_middleware(
                self._mock_data_middleware[db_name],
                priority=9
            )
        super(RestModelTestMixin, self).setUp()

    def tearDown(self):
        # remove all middelwaress
        for db_name, middleware in self._mock_data_middleware.items():
            connections[db_name].connection.pop_middleware(middleware)

    @contextmanager
    def mock_api(self, url, result, params=None, using=None):
        """
        assert that the eather one of the api have executed a query, with optionnaly the given params,
        the query was made `count` times and by the given API

        :param str|None url: the url in which the query was made. if None, all url will be count
        :param params: the params that must be present in the query
        :param str using: the name of connection to use
        :param dict|None|list result: the temporary result to provide for the given time
        """
        connection = connections[using or get_default_api_database(settings.DATABASES)]
        mocked = {
            url: [{
                'filter': params or {},
                'data': result
            }]
        }
        middleware = MockDataApiMiddleware(mocked, not_found=not_found_continue)
        cursor = connection.cursor()
        try:

            cursor.push_middleware(middleware, priority=7)
            yield
        finally:
            cursor.pop_middleware(middleware)


class RestModelTestCase(RestModelTestMixin, TestCase):
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging
import re

from django.conf import settings
from django.db.backends.base.creation import BaseDatabaseCreation

from rest_models.backend.connexion import LocalApiAdapter

logger = logging.getLogger(__name__)


def do_nothing(*args, **kwargs):
    pass


class DatabaseCreation(BaseDatabaseCreation):
    """
    dont create anything in this database: its an API noob
    """
    sql_destroy_indexes_for_field = do_nothing
    sql_remove_table_constraints = do_nothing
    sql_indexes_for_fields = do_nothing
    serialize_db_to_string = do_nothing
    sql_destroy_model = do_nothing
    sql_destroy_indexes_for_fields = do_nothing
    sql_destroy_indexes_for_model = do_nothing
    sql_indexes_for_field = do_nothing
    sql_table_creation_suffix = do_nothing
    sql_for_inline_foreign_key_references = do_nothing
    deserialize_db_from_string = do_nothing
    sql_for_pending_references = do_nothing
    sql_indexes_for_model = do_nothing
    sql_create_model = do_nothing
    destroy_test_db = do_nothing

    def create_test_db(self, verbosity=1, autoclobber=False, serialize=True, keepdb=False):
        """
        Creates a test database, prompting the user for confirmation if the
        database already exists. Returns the name of the test database created.
        """
        # Don't import django.core.management if it isn't needed.
        test_override = self.connection.settings_dict.get('TEST', {})
        if any(test_override.values()):
            self.connection.settings_dict.update(test_override)
        elif not self.connection.alias.startswith('TEST_'):
            test_database_name = self._get_test_db_name()
            settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
            self.connection.settings_dict["NAME"] = test_database_name
        return self.connection.settings_dict["NAME"]

    def _get_test_db_name(self):
        """
        Internal implementation - returns the name of the test DB that will be
        created. Only useful when called from create_test_db() and
        _create_test_db() and when no external munging is done with the 'NAME'
        settings.
        """
        test_alias = 'TEST_'+self.connection.alias
        if settings.DATABASES.get(test_alias):
            return settings.DATABASES[test_alias]['NAME']
        name = self.connection.settings_dict['NAME']
        return re.sub('https?://[^/]+/', LocalApiAdapter.SPECIAL_URL + "/", name, count=1)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

logger = logging.getLogger(__name__)


def message_from_response(response):
    return "[%d]%s" % (
        response.status_code,
        response.text if '<!DOCTYPE html>' not in response.text[:30] else response.reason
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import unittest

import rest_models

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = rest_models.__VERSION__

if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = 'testsettings'

if sys.argv[-1] == 'publish':
    # os.system('cd docs && make html')
    os.system('python setup.py sdist bdist_wheel upload')
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

if sys.argv[-1] == 'doc':
    os.system('cd docs && make html')
    sys.exit()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme_file:
    readme = readme_file.read()

with open(os.path.join(os.path.dirname(__file__), 'test_requirements.txt')) as requirements_file:
    tests_require = requirements_file.readlines()

setup(
    name='django-rest-models',
    version=version,
    description="""django Fake ORM model that query an RestAPI instead of a database — """,
    long_description=readme,
    author='Darius BERNARD',
    author_email='darius@yupeek.com',
    url='https://github.com/Yupeek/django-rest-models',
    tests_require=tests_require,
    install_requires=[
        'requests',
    ],
    packages=[
        'rest_models',
        'rest_models.backend',

    ],
    include_package_data=True,
    license="GNU GENERAL PUBLIC LICENSE",
    zip_safe=False,
    keywords='django rest models API ORM',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Environment :: Web Environment',
        'Framework :: Django',
    ],
)





from django.db.backends.base.schema import BaseDatabaseSchemaEditor


def do_nothing(*args, **kwargs):  # pragma: no cover
    pass


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    alter_index_together = do_nothing
    column_sql = do_nothing
    alter_db_tablespace = do_nothing
    execute = do_nothing
    alter_db_table = do_nothing
    create_model = do_nothing
    skip_default = do_nothing
    alter_field = do_nothing
    alter_unique_together = do_nothing
    remove_field = do_nothing
    effective_default = do_nothing
    add_field = do_nothing
    prepare_default = do_nothing
    delete_model = do_nothing
    quote_name = do_nothing
    quote_value = do_nothing



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging

logger = logging.getLogger(__name__)


class FakeApiResponse(object):
    def __init__(self, data, status_code):
        self.data = data
        self.status_code = status_code

    def json(self):
        return self.data

    @property
    def text(self):
        try:
            return json.dumps(self.data)
        except TypeError:
            return repr(self.data)


class ApiMiddleware(object):
    """
    a base class to implemente a middleware that will
    interact with a api query/response
    """

    def process_request(self, params, requestid, connection):
        """
        process the request. if return other than None, it will be the result of the request

        :param dict params: the params given (and modified by previous middleware) for the _make_query.
                            the params can be updated at will by side-effect.
                    params contain
                        - verb: the verb to execut the request (post, get, put)
                        - url: the url in which the query will be executed
                        - data: the data given to the query to post,put, etc

        :param int requestid: the id of the current request done by this connection
        :param connection: the connection used for this query
        :return: the response if there is no need to pursue the query.
        """
        return None

    def process_response(self, params, response, requestid):
        """
        process the response from a previous query. MUST return a response (result)

        :param params: the params finaly given to query the api. same format as for process_request
        :param response:  the response, either the original one or modifier by preceding middleware
        :param int requestid: the id of the current request done by this connection
        :return:
        """
        return response

    @staticmethod
    def make_response(data, status_code):
        """
        helper to make a response (returned by process_response or process_request) with given data

        :param data: the data in the response (will be encoded in json to be compatible)
        :param status_code: the status code of the response
        :return: a FakeApiResponse that contains raw data
        """
        return FakeApiResponse(data, status_code)

    def empty_response(self):
        """
        shortcut to return a response with 204 status code and no data, which will be

        :return: a FakeApiResponse with no data and 204 status code
        """
        return self.make_response(data=None, status_code=204)

    def data_response(self, data, status_code=None):
        """
        shortcut to return a response with 200 and data

        :param data: the data to insert in the response
        :return: a FakeApiResponse with the given data
        """
        return self.make_response(data=data, status_code=status_code or 200)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

from django.apps import apps
from django.core.checks import Error, Tags, Warning, register

logger = logging.getLogger(__name__)


def api_struct_check(app_configs, **kwargs):
    from rest_models.backend.compiler import get_resource_path  # NOQA
    from rest_models.router import RestModelRouter  # NOQA

    errors = []

    all_models = []
    if app_configs is None:
        all_models.extend(apps.get_models())
    else:
        for app_config in app_configs:
            all_models.extend(app_config.get_models())
    router = RestModelRouter()
    models = ((router.get_api_connexion(model).cursor(), model)
              for model in all_models if router.is_api_model(model))

    for db, rest_model in models:
        url = get_resource_path(rest_model)
        res = db.options(url)
        if res.status_code != 200:
            errors.append(
                Error(
                    'the remote api does not respond to us. OPTIONS %s%s => %s' % (db.url, url, res.status_code),
                    hint='check the url for the remote api or the resource_path',
                    obj=rest_model,
                    id='rest_models.E001'
                )
            )
            continue
        options = res.json()
        missings = {
                       'include[]', 'exclude[]', 'filter{}', 'page', 'per_page', 'sort[]'
                   } - set(options.get("features", []))
        if missings:
            errors.append(
                Error(
                    'the remote api does not support the folowing features: %s' % missings,
                    hint='is the api on %s/%s running with dynamic-rest ?' % (db.url, url),
                    obj=rest_model,
                    id='rest_models.E002'
                )
            )
            continue
        for field in rest_model._meta.get_fields():
            if field.is_relation:
                if router.is_api_model(field.related_model):
                    if field.name not in options['properties']:
                        errors.append(
                            Error(
                                'the field %s.%s in not present on the remote serializer' % (
                                    rest_model.__name__, field.name
                                ),
                                obj="%s.%s" % (rest_model.__name__, field.name),
                                hint='check if the serializer on %s/%s has a field "%s"' % (db.url, url, field.name),
                                id='rest_models.E003'
                            )
                        )
                    else:
                        type_is_many = options['properties'][field.name]['type'] == 'many'
                        type_is_one = options['properties'][field.name]['type'] == 'one'
                        if (type_is_many and not (field.one_to_many or field.many_to_many) or
                                type_is_one and not (field.one_to_one or field.many_to_one)):

                            errors.append(
                                Error(
                                    'the field %s.%s many does not match the api' % (
                                        rest_model.__name__, field.name
                                    ),
                                    obj="%s.%s" % (rest_model.__name__, field.name),
                                    hint='check if the serializer at %s%s have a Serializer.many '
                                         'value corresponding to the local model %s' % (db.url, url, field.name),
                                    id='rest_models.E005'
                                )
                            )

                        choice_count = len(options['properties'][field.name].get('choices', []))
                        if choice_count > 100:
                            errors.append(
                                Warning(
                                    'the field %s.%s has many choices values (%s) in OPTIONS '
                                    'and it slow down the check' % (
                                        rest_model.__name__, field.name, choice_count
                                    ),
                                    obj="%s.%s" % (rest_model.__name__, field.name),
                                    hint='check if the serializer at %s%s provide '
                                         'a choices with less values for %s' % (
                                        db.url, url, field.name),
                                    id='rest_models.W001'
                                )
                            )

            elif field.name not in options['properties']:
                errors.append(
                    Error(
                        'the field %s.%s in not present on the remote serializer' % (rest_model.__name__, field.name),
                        hint='check if the serializer on %s%s has a field "%s"' % (db.url, url, field.name),
                        id='rest_models.E006'
                    )
                )
    return errors


def register_checks():
    register(api_struct_check, Tags.models)
